/* 
   ssl.c - TLSv1 cryptography support

   Copyright (C) Szymon Acedanski 2003-2004
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003-2004 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#include "ggrelay.h"

static SSL_CTX* ssl_svr_ctx;
static pthread_mutex_t* ssl_locks;

static void gg_ssl_reperr(void)
{
	char buf[256];
	int err;
	while ((err = ERR_get_error()))
	{
		ERR_error_string_n(err, buf, sizeof(buf));
		gg_condlog("ssl error: %s", buf);
	}
}

static void gg_ssl_locking_helper(int mode, int n, const char* file, int line)
{
	if (mode & CRYPTO_LOCK)
		pthread_mutex_lock(ssl_locks+n);
	else
		pthread_mutex_unlock(ssl_locks+n);
}

int gg_ssl_init(const char* key, const char* cert)
{
	int i;
	int n;
	
	SSL_load_error_strings();
	SSL_library_init();

	n = CRYPTO_num_locks();
	if (!(ssl_locks = (pthread_mutex_t*)malloc(n*sizeof(pthread_mutex_t))))
	{
		gg_error("fatal: error allocating locks for OpenSSL");
		return 1;
	}
	for (i=0; i<n; i++)
		pthread_mutex_init(ssl_locks+i, NULL);
	CRYPTO_set_locking_callback(gg_ssl_locking_helper);
	
	if (!RAND_status())
	{
		char rdata[1024];
		struct
		{
			time_t time;
			void *ptr;
		} rstruct;

		time(&rstruct.time);
		rstruct.ptr = &rstruct;			

		RAND_seed(rdata, sizeof(rdata));
		RAND_seed(&rstruct, sizeof(rstruct));
	}

	if (!(ssl_svr_ctx = SSL_CTX_new(SSLv23_server_method())))
		goto failure;
	if (!SSL_CTX_use_RSAPrivateKey_file(ssl_svr_ctx, key, SSL_FILETYPE_PEM))
		goto failure;
	if (!SSL_CTX_use_certificate_file(ssl_svr_ctx, cert, SSL_FILETYPE_PEM))
		goto failure;
	if (!SSL_CTX_check_private_key(ssl_svr_ctx))
		goto failure;
	SSL_CTX_set_mode(ssl_svr_ctx,
			SSL_MODE_ENABLE_PARTIAL_WRITE | SSL_MODE_AUTO_RETRY);
	/* It violates TLS protocol, but is simple. */
	SSL_CTX_set_quiet_shutdown(ssl_svr_ctx, 1);
	return 0;
	
failure:
	gg_ssl_reperr();
	gg_error("fatal: could not initialize SSL");
	return 1;
}

int gg_ssl_accept_connection(struct gg_fd* f)
{
	if (!(f->ssl = SSL_new(ssl_svr_ctx)))
		goto failure;
	if (!SSL_set_fd(f->ssl, f->fd))
		goto freessl;
	DEBUGP("accepting ...");
	if (SSL_accept(f->ssl) <= 0)
		goto freessl;
	DEBUGP("accepted");
	return 0;

freessl:
	SSL_free(f->ssl);
	f->ssl = NULL;
failure:
	DEBUGP("failed");
	gg_ssl_reperr();
	return 1;
}

int gg_ssl_write(struct gg_fd* f)
{
	int err = SSL_write(f->ssl, f->twin->buffer, f->twin->writeoffset);
	int sslerr;

	f->flags &= ~(GG_FLAG_READFORWRITING | GG_FLAG_WAITFORME);
	if (err >= 0) return err;
	sslerr = SSL_get_error(f->ssl, err);
	DEBUGP("fd=%d err=%d sslerr=%d", f->fd, err, sslerr);
	switch (sslerr)
	{
		case SSL_ERROR_WANT_READ:
			f->flags |= GG_FLAG_READFORWRITING;
		case SSL_ERROR_WANT_WRITE:
			f->flags |= GG_FLAG_WAITFORME;
			return -2;
		case SSL_ERROR_SYSCALL:
			return err;
		case SSL_ERROR_SSL:
			gg_condlog("SSL error on fd %d:", f->fd);
			gg_ssl_reperr();
	}
	return -1;
}

int gg_ssl_read(struct gg_fd* f)
{
	int err = SSL_read(f->ssl, f->buffer + f->readoffset,
			f->bufsize - f->readoffset);
	int sslerr;

	f->flags &= ~(GG_FLAG_WRITEFORREADING | GG_FLAG_WAITFORME);
	if (err >= 0) return err;
	sslerr = SSL_get_error(f->ssl, err);
	DEBUGP("fd=%d err=%d sslerr=%d", f->fd, err, sslerr);
	switch (sslerr)
	{
		case SSL_ERROR_WANT_WRITE:
			f->flags |= GG_FLAG_WRITEFORREADING;
		case SSL_ERROR_WANT_READ:
			f->flags |= GG_FLAG_WAITFORME;
			return -2;
		case SSL_ERROR_SYSCALL:
			return err;
		case SSL_ERROR_SSL:
			gg_condlog("SSL error on fd %d:", f->fd);
			gg_ssl_reperr();
	}
	return -1;
}

void gg_ssl_shutdown(struct gg_fd* f, int mode)
{
}

void gg_ssl_close(struct gg_fd* f)
{
	if (f->ssl)
	{
		SSL_shutdown(f->ssl);
		SSL_free(f->ssl);
		f->ssl = NULL;
	}
	shutdown(f->fd, SHUT_RDWR);
	gg_close_and_unmap_fd(f);
}
